@@ -1,10 +1,13 @@
+#include <config.h>
+
 #include <apt-pkg/error.h>
 #include <apt-pkg/acquire-method.h>
 #include <apt-pkg/strutl.h>
-#include <apti18n.h>
+#include <apt-pkg/fileutl.h>
+#include <apt-pkg/indexcopy.h>
+#include <apt-pkg/configuration.h>
+#include <apt-pkg/gpgv.h>
 
-#include <sys/stat.h>
-#include <unistd.h>
 #include <utime.h>
 #include <stdio.h>
 #include <fcntl.h>
@@ -12,6 +15,12 @@
 #include <sys/wait.h>
 #include <iostream>
 #include <sstream>
+#include <vector>
+
+#include <apti18n.h>
+
+using std::string;
+using std::vector;
 
 #define GNUPGPREFIX "[GNUPG:]"
 #define GNUPGBADSIG "[GNUPG:] BADSIG"
@@ -45,100 +54,25 @@ string GPGVMethod::VerifyGetSigners(const char *file, const char *outfile,
 					 vector<string> &WorthlessSigners,
 					 vector<string> &NoPubKeySigners)
 {
-   // setup a (empty) stringstream for formating the return value
-   std::stringstream ret;
-   ret.str("");
+   bool const Debug = _config->FindB("Debug::Acquire::gpgv", false);
+
+   if (Debug == true)
+      std::clog << "inside VerifyGetSigners" << std::endl;
 
-   if (_config->FindB("Debug::Acquire::gpgv", false))
-   {
-      std::cerr << "inside VerifyGetSigners" << std::endl;
-   }
-   pid_t pid;
    int fd[2];
-   FILE *pipein;
-   int status;
-   struct stat buff;
-   string gpgvpath = _config->Find("Dir::Bin::gpg", "/usr/bin/gpgv");
-   string pubringpath = _config->Find("APT::GPGV::TrustedKeyring", "/etc/apt/trusted.gpg");
-   if (_config->FindB("Debug::Acquire::gpgv", false))
-   {
-      std::cerr << "gpgv path: " << gpgvpath << std::endl;
-      std::cerr << "Keyring path: " << pubringpath << std::endl;
-   }
 
-   if (stat(pubringpath.c_str(), &buff) != 0) 
-   {
-      ioprintf(ret, _("Couldn't access keyring: '%s'"), strerror(errno)); 
-      return ret.str();
-   }
    if (pipe(fd) < 0)
-   {
       return "Couldn't create pipe";
-   }
 
-   pid = fork();
+   pid_t pid = fork();
    if (pid < 0)
-   {
       return string("Couldn't spawn new process") + strerror(errno);
-   }
    else if (pid == 0)
-   {
-      const char *Args[400];
-      unsigned int i = 0;
-
-      Args[i++] = gpgvpath.c_str();
-      Args[i++] = "--status-fd";
-      Args[i++] = "3";
-      Args[i++] = "--ignore-time-conflict";
-      Args[i++] = "--keyring";
-      Args[i++] = pubringpath.c_str();
-
-      Configuration::Item const *Opts;
-      Opts = _config->Tree("Acquire::gpgv::Options");
-      if (Opts != 0)
-      {
-         Opts = Opts->Child;
-	 for (; Opts != 0; Opts = Opts->Next)
-         {
-            if (Opts->Value.empty() == true)
-               continue;
-            Args[i++] = Opts->Value.c_str();
-	    if(i >= 395) { 
-	       std::cerr << _("E: Argument list from Acquire::gpgv::Options too long. Exiting.") << std::endl;
-	       exit(111);
-	    }
-         }
-      }
-      Args[i++] = file;
-      Args[i++] = outfile;
-      Args[i++] = NULL;
-
-      if (_config->FindB("Debug::Acquire::gpgv", false))
-      {
-         std::cerr << "Preparing to exec: " << gpgvpath;
-	 for(unsigned int j=0;Args[j] != NULL; j++)
-	    std::cerr << " " << Args[j];
-	 std::cerr << std::endl;
-      }
-      int nullfd = open("/dev/null", O_RDONLY);
-      close(fd[0]);
-      // Redirect output to /dev/null; we read from the status fd
-      dup2(nullfd, STDOUT_FILENO); 
-      dup2(nullfd, STDERR_FILENO); 
-      // Redirect the pipe to the status fd (3)
-      dup2(fd[1], 3);
-
-      putenv((char *)"LANG=");
-      putenv((char *)"LC_ALL=");
-      putenv((char *)"LC_MESSAGES=");
-      execvp(gpgvpath.c_str(), (char **)Args);
-             
-      exit(111);
-   }
+      ExecGPGV(outfile, file, 3, fd);
    close(fd[1]);
 
-   pipein = fdopen(fd[0], "r"); 
-   
+   FILE *pipein = fdopen(fd[0], "r");
+
    // Loop over the output of gpgv, and check the signatures.
    size_t buffersize = 64;
    char *buffer = (char *) malloc(buffersize);
@@ -150,8 +84,16 @@ string GPGVMethod::VerifyGetSigners(const char *file, const char *outfile,
       // Read a line.  Sigh.
       while ((c = getc(pipein)) != EOF && c != '\n')
       {
-         if (bufferoff == buffersize)
-            buffer = (char *) realloc(buffer, buffersize *= 2);
+	 if (bufferoff == buffersize)
+	 {
+	    char* newBuffer = (char *) realloc(buffer, buffersize *= 2);
+	    if (newBuffer == NULL)
+	    {
+	       free(buffer);
+	       return "Couldn't allocate a buffer big enough for reading";
+	    }
+	    buffer = newBuffer;
+	 }
          *(buffer+bufferoff) = c;
          bufferoff++;
       }
@@ -159,8 +101,8 @@ string GPGVMethod::VerifyGetSigners(const char *file, const char *outfile,
          break;
       *(buffer+bufferoff) = '\0';
       bufferoff = 0;
-      if (_config->FindB("Debug::Acquire::gpgv", false))
-         std::cerr << "Read: " << buffer << std::endl;
+      if (Debug == true)
+         std::clog << "Read: " << buffer << std::endl;
 
       // Push the data into three separate vectors, which
       // we later concatenate.  They're kept separate so
@@ -168,33 +110,33 @@ string GPGVMethod::VerifyGetSigners(const char *file, const char *outfile,
       // it will be better.
       if (strncmp(buffer, GNUPGBADSIG, sizeof(GNUPGBADSIG)-1) == 0)
       {
-         if (_config->FindB("Debug::Acquire::gpgv", false))
-            std::cerr << "Got BADSIG! " << std::endl;
+         if (Debug == true)
+            std::clog << "Got BADSIG! " << std::endl;
          BadSigners.push_back(string(buffer+sizeof(GNUPGPREFIX)));
       }
       
       if (strncmp(buffer, GNUPGNOPUBKEY, sizeof(GNUPGNOPUBKEY)-1) == 0)
       {
-         if (_config->FindB("Debug::Acquire::gpgv", false))
-            std::cerr << "Got NO_PUBKEY " << std::endl;
+         if (Debug == true)
+            std::clog << "Got NO_PUBKEY " << std::endl;
          NoPubKeySigners.push_back(string(buffer+sizeof(GNUPGPREFIX)));
       }
       if (strncmp(buffer, GNUPGNODATA, sizeof(GNUPGBADSIG)-1) == 0)
       {
-         if (_config->FindB("Debug::Acquire::gpgv", false))
-            std::cerr << "Got NODATA! " << std::endl;
+         if (Debug == true)
+            std::clog << "Got NODATA! " << std::endl;
          BadSigners.push_back(string(buffer+sizeof(GNUPGPREFIX)));
       }
       if (strncmp(buffer, GNUPGKEYEXPIRED, sizeof(GNUPGKEYEXPIRED)-1) == 0)
       {
-         if (_config->FindB("Debug::Acquire::gpgv", false))
-            std::cerr << "Got KEYEXPIRED! " << std::endl;
+         if (Debug == true)
+            std::clog << "Got KEYEXPIRED! " << std::endl;
          WorthlessSigners.push_back(string(buffer+sizeof(GNUPGPREFIX)));
       }
       if (strncmp(buffer, GNUPGREVKEYSIG, sizeof(GNUPGREVKEYSIG)-1) == 0)
       {
-         if (_config->FindB("Debug::Acquire::gpgv", false))
-            std::cerr << "Got REVKEYSIG! " << std::endl;
+         if (Debug == true)
+            std::clog << "Got REVKEYSIG! " << std::endl;
          WorthlessSigners.push_back(string(buffer+sizeof(GNUPGPREFIX)));
       }
       if (strncmp(buffer, GNUPGGOODSIG, sizeof(GNUPGGOODSIG)-1) == 0)
@@ -204,17 +146,18 @@ string GPGVMethod::VerifyGetSigners(const char *file, const char *outfile,
          while (*p && isxdigit(*p)) 
             p++;
          *p = 0;
-         if (_config->FindB("Debug::Acquire::gpgv", false))
-            std::cerr << "Got GOODSIG, key ID:" << sig << std::endl;
+         if (Debug == true)
+            std::clog << "Got GOODSIG, key ID:" << sig << std::endl;
          GoodSigners.push_back(string(sig));
       }
    }
    fclose(pipein);
 
+   int status;
    waitpid(pid, &status, 0);
-   if (_config->FindB("Debug::Acquire::gpgv", false))
+   if (Debug == true)
    {
-      std::cerr << "gpgv exited\n";
+      std::clog << "gpgv exited\n";
    }
    
    if (WEXITSTATUS(status) == 0)
@@ -224,18 +167,19 @@ string GPGVMethod::VerifyGetSigners(const char *file, const char *outfile,
       return "";
    }
    else if (WEXITSTATUS(status) == 1)
-   {
       return _("At least one invalid signature was encountered.");
-   }
    else if (WEXITSTATUS(status) == 111)
+      return _("Could not execute 'gpgv' to verify signature (is gpgv installed?)");
+   else if (WEXITSTATUS(status) == 112)
    {
-      ioprintf(ret, _("Could not execute '%s' to verify signature (is gnupg installed?)"), gpgvpath.c_str());
-      return ret.str();
+      // acquire system checks for "NODATA" to generate GPG errors (the others are only warnings)
+      std::string errmsg;
+      //TRANSLATORS: %s is a single techy word like 'NODATA'
+      strprintf(errmsg, _("Clearsigned file isn't valid, got '%s' (does the network require authentication?)"), "NODATA");
+      return errmsg;
    }
    else
-   {
       return _("Unknown error executing gpgv");
-   }
 }
 
 bool GPGVMethod::Fetch(FetchItem *Itm)
@@ -270,21 +214,21 @@ bool GPGVMethod::Fetch(FetchItem *Itm)
          {
             errmsg += _("The following signatures were invalid:\n");
             for (vector<string>::iterator I = BadSigners.begin();
-		 I != BadSigners.end(); I++)
+		 I != BadSigners.end(); ++I)
                errmsg += (*I + "\n");
          }
          if (!WorthlessSigners.empty())
          {
             errmsg += _("The following signatures were invalid:\n");
             for (vector<string>::iterator I = WorthlessSigners.begin();
-		 I != WorthlessSigners.end(); I++)
+		 I != WorthlessSigners.end(); ++I)
                errmsg += (*I + "\n");
          }
          if (!NoPubKeySigners.empty())
          {
              errmsg += _("The following signatures couldn't be verified because the public key is not available:\n");
             for (vector<string>::iterator I = NoPubKeySigners.begin();
-		 I != NoPubKeySigners.end(); I++)
+		 I != NoPubKeySigners.end(); ++I)
                errmsg += (*I + "\n");
          }
       }
@@ -292,7 +236,7 @@ bool GPGVMethod::Fetch(FetchItem *Itm)
       // least one bad signature. good signatures and NoPubKey signatures
       // happen easily when a file is signed with multiple signatures
       if(GoodSigners.empty() or !BadSigners.empty())
-      	 return _error->Error(errmsg.c_str());
+      	 return _error->Error("%s", errmsg.c_str());
    }
       
    // Just pass the raw output up, because passing it as a real data
@@ -305,7 +249,7 @@ bool GPGVMethod::Fetch(FetchItem *Itm)
 
    if (_config->FindB("Debug::Acquire::gpgv", false))
    {
-      std::cerr << "gpgv succeeded\n";
+      std::clog << "gpgv succeeded\n";
    }
 
    return true;
